require "thor"
require "pathname"
require "json"
require "bbc/cosmos/tools/config"

module BBC
  module Cosmos
    module Tools
      module Commands
        class Base < Thor
          attr_accessor :config

          include Thor::Actions

          package_name "BBC Cosmos config tool"

          class_option :key_path, :type => :string, :default => ENV["BBC_COSMOS_TOOLS_CERT"], :desc => "The path to your dev cert, by default the value of the env variable $BBC_COSMOS_TOOLS_CERT is used"
          class_option :env, :type => :string, :default => "int", :desc => "The environment to target"
          class_option :base, :type => :string, :desc => "The base path for the configs repo"
          class_option :project, :type => :string, :default => ENV["BBC_COSMOS_TOOLS_PROJECT"], :desc => "The name of the project, by default the value of the env variable $BBC_COSMOS_TOOLS_PROJECT is used"

          def initialize(args = [], local_options = {}, config = {})
            super(args, local_options, config)
            setup_config
          end

          protected

          def get_key_value(key, value)
            "#{set_color("#{key}:", :cyan)} #{set_color(value, :white)}"
          end

          def banner(component = nil)
            output = component.nil? ? "\n" : "\n#{get_key_value('Component', component)}\n"
            options.reduce(output) do |string, (key, value)|
              string += "#{get_key_value(key.capitalize.gsub(/_/, ' '), value)}\n"
            end
          end

          def components(options, component = nil)
            components = component.nil? ? config.components_for(options[:tags]).keys : [component]
          end

          def api_error(response, die = true)
            error_message = JSON.parse(response.body)["message"]
            error(error_message, die)
          end

          def build_table(headers, data)
            [].tap do |table|
              table << headers.reduce([]) do |array, header|
                array << set_color(header, :cyan)
              end

              data.each do |items|
                table << items.map { |item| set_color(item, :white) }
              end
            end
          end

          def error(message, die = true)
            say "#{message}\n", :red
            exit 1 if die
          end

          private

          def setup_config
            config_base = !options[:base].nil? && File.directory?(options[:base]) ? Pathname.new(options[:base]) : Pathname.pwd.join(".")
            fail("Config directory: #{config_base} doesn't exist") unless File.directory?(config_base)
            @config = BBC::Cosmos::Tools::Config.new(
              config_base,
              options[:project],
              options[:env]
            )
          end
        end
      end
    end
  end
end
